import { Construct, IConstruct } from 'constructs';
import { Chart } from './chart';
import { JsonPatch } from './json-patch';
import { ApiObjectMetadata, ApiObjectMetadataDefinition } from './metadata';
/**
 * (experimental) Options for defining API objects.
 *
 * @experimental
 */
export interface ApiObjectProps {
    /**
     * (experimental) Object metadata.
     *
     * If `name` is not specified, an app-unique name will be allocated by the
     * framework based on the path of the construct within thes construct tree.
     *
     * @experimental
     */
    readonly metadata?: ApiObjectMetadata;
    /**
     * (experimental) API version.
     *
     * @experimental
     */
    readonly apiVersion: string;
    /**
     * (experimental) Resource kind.
     *
     * @experimental
     */
    readonly kind: string;
    /**
     * Additional attributes for this API object.
     */
    readonly [key: string]: any;
}
/**
 * @experimental
 */
export declare class ApiObject extends Construct {
    private readonly props;
    /**
     * (experimental) Returns the `ApiObject` named `Resource` which is a child of the given construct.
     *
     * If `c` is an `ApiObject`, it is returned directly. Throws an
     * exception if the construct does not have a child named `Default` _or_ if
     * this child is not an `ApiObject`.
     *
     * @param c The higher-level construct.
     * @experimental
     */
    static of(c: IConstruct): ApiObject;
    /**
     * (experimental) The name of the API object.
     *
     * If a name is specified in `metadata.name` this will be the name returned.
     * Otherwise, a name will be generated by calling
     * `Chart.of(this).generatedObjectName(this)`, which by default uses the
     * construct path to generate a DNS-compatible name for the resource.
     *
     * @experimental
     */
    readonly name: string;
    /**
     * (experimental) The object's API version (e.g. `authorization.k8s.io/v1`).
     *
     * @experimental
     */
    readonly apiVersion: string;
    /**
     * (experimental) The group portion of the API version (e.g. `authorization.k8s.io`).
     *
     * @experimental
     */
    readonly apiGroup: string;
    /**
     * (experimental) The object kind.
     *
     * @experimental
     */
    readonly kind: string;
    /**
     * (experimental) The chart in which this object is defined.
     *
     * @experimental
     */
    readonly chart: Chart;
    /**
     * (experimental) Metadata associated with this API object.
     *
     * @experimental
     */
    readonly metadata: ApiObjectMetadataDefinition;
    /**
     * A set of JSON patch operations to apply to the document after synthesis.
     */
    private readonly patches;
    /**
     * (experimental) Defines an API object.
     *
     * @param scope the construct scope.
     * @param id namespace.
     * @param props options.
     * @experimental
     */
    constructor(scope: Construct, id: string, props: ApiObjectProps);
    /**
     * (experimental) Create a dependency between this ApiObject and other constructs.
     *
     * These can be other ApiObjects, Charts, or custom.
     *
     * @param dependencies the dependencies to add.
     * @experimental
     */
    addDependency(...dependencies: IConstruct[]): void;
    /**
     * (experimental) Applies a set of RFC-6902 JSON-Patch operations to the manifest synthesized for this API object.
     *
     * @param ops The JSON-Patch operations to apply.
     * @experimental
     * @example
     *
     *   kubePod.addJsonPatch(JsonPatch.replace('/spec/enableServiceLinks', true));
     */
    addJsonPatch(...ops: JsonPatch[]): void;
    /**
     * (experimental) Renders the object to Kubernetes JSON.
     *
     * @experimental
     */
    toJson(): any;
}
